from privacyscanner.scanmodules.testsslsh.base import TestsslshScanModuleBase
from privacyscanner.scanmodules.testsslsh.scanner import Parameter


class TestsslshHttpsScanModule(TestsslshScanModuleBase):
    name = 'testssl_https'
    required_keys = ['final_url', 'https', 'testssl_https']
    target_type = 'https'
    target_parameters = []

    def _get_host(self, result):
        host_url = result['final_url']
        # An HTTP site might have an HTTPS version too, but does not redirect
        # to it. In this case, we still want to scan the HTTPS version.
        if host_url.startswith('http://'):
            host_url = 'https://' + host_url[len('http://'):]
        return host_url

    def _can_run(self, result):
        return 'https' in result and result['https']['has_tls']


class TestsslshMailScanModule(TestsslshScanModuleBase):
    name = 'testssl_mail'
    required_keys = ['mail', 'testssl_mail']
    target_type = 'mail'
    target_parameters = [Parameter.STARTTLS, 'smtp']

    def _get_host(self, result):
        return result['mail']['domain'] + ':25'

    def _can_run(self, result):
        # If the mail host is not reachable, the has_starttls key is
        # intentionally missing.
        # If the mail host is reachable, but we could not successfully
        # perform an EHLO to switch to STARTTLS later on, we skip
        # scanning too.
        return 'mail' in result and bool(result['mail'].get('has_starttls'))



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class CookiesExtractor(Extractor):
    def extract_information(self):
        cookies = self.page.tab.Network.getAllCookies()['cookies']
        timestamp = int(self.page.scan_start.timestamp())
        for cookie in cookies:
            if cookie['session'] or cookie['expires'] is None:
                cookie['lifetime'] = -1
            else:
                cookie['lifetime'] = cookie['expires'] - timestamp
        self.result['cookies'] = cookies



from pathlib import Path

from privacyscanner.filehandlers import NoOpFileHandler
from privacyscanner.result import Result
from privacyscanner.scanmodules import ScanModule
from privacyscanner.scanmodules.chromedevtools.chromescan import ChromeScan, find_chrome_executable
from privacyscanner.scanmodules.chromedevtools.extractors import FinalUrlExtractor, \
    GoogleAnalyticsExtractor, CookiesExtractor, RequestsExtractor, RedirectChainExtractor, \
    TLSDetailsExtractor, CertificateExtractor, ThirdPartyExtractor, InsecureContentExtractor, \
    FailedRequestsExtractor, SecurityHeadersExtractor, TrackerDetectExtractor, \
    CookieStatsExtractor, JavaScriptLibsExtractor, ScreenshotExtractor, ImprintExtractor, \
    HSTSPreloadExtractor, FingerprintingExtractor
from privacyscanner.scanmodules.chromedevtools.utils import TLDEXTRACT_CACHE_FILE, parse_domain
from privacyscanner.utils import file_is_outdated, set_default_options, calculate_jaccard_index


EXTRACTOR_CLASSES = [FinalUrlExtractor, RedirectChainExtractor, GoogleAnalyticsExtractor,
                     CookiesExtractor, RequestsExtractor, TLSDetailsExtractor,
                     CertificateExtractor, ThirdPartyExtractor, InsecureContentExtractor,
                     FailedRequestsExtractor, SecurityHeadersExtractor, TrackerDetectExtractor,
                     CookieStatsExtractor, JavaScriptLibsExtractor, ScreenshotExtractor,
                     ImprintExtractor, HSTSPreloadExtractor, FingerprintingExtractor]

EXTRACTOR_CLASSES_HTTPS_RUN = [FinalUrlExtractor, TLSDetailsExtractor, CertificateExtractor,
                               InsecureContentExtractor, SecurityHeadersExtractor,
                               HSTSPreloadExtractor]


class ChromeDevtoolsScanModule(ScanModule):
    name = 'chromedevtools'
    dependencies = []
    required_keys = ['site_url']

    def __init__(self, options):
        if 'chrome_executable' not in options:
            options['chrome_executable'] = find_chrome_executable()
        set_default_options(options, {
            'disable_javascript': False,
            'https_same_content_threshold': 0.9
        })
        super().__init__(options)
        cache_file = self.options['storage_path'] / TLDEXTRACT_CACHE_FILE
        parse_domain.cache_file = str(cache_file)

    def scan_site(self, result, meta):
        chrome_scan = ChromeScan(EXTRACTOR_CLASSES)
        debugging_port = self.options.get('start_port', 9222) + meta.worker_id
        content = chrome_scan.scan(result, self.logger, self.options, meta, debugging_port)
        if not result['reachable']:
            return
        result['https']['same_content'] = None
        result['https']['same_content_score'] = None
        if result['site_url'].startswith('http://') and not result['https']['redirects_secure']:
            # Lets do another scan with https but with limited extractors.
            # We use this to annotate the http result with TLS details and
            # insecure content details if there is not redirect to https
            site_url = 'https://' + result['site_url'][len('http://'):]
            extra_result = Result({'site_url': site_url}, NoOpFileHandler())
            chrome_scan = ChromeScan(EXTRACTOR_CLASSES_HTTPS_RUN)
            https_content = chrome_scan.scan(extra_result, self.logger, self.options, meta,
                                             debugging_port)
            if not extra_result['reachable']:
                return
            similarity = calculate_jaccard_index(content, https_content)
            same_content = similarity >= self.options['https_same_content_threshold']
            if same_content:
                result['insecure_content'] = extra_result['insecure_content']
                result['https'] = extra_result['https']
                result['https']['redirects_secure'] = False
            result['https']['same_content_score'] = similarity
            result['https']['same_content'] = same_content

    def update_dependencies(self):
        max_age = 14 * 24 * 3600
        cache_file = Path(parse_domain.cache_file)
        cache_file.parent.mkdir(parents=True, exist_ok=True)
        if file_is_outdated(cache_file, max_age):
            parse_domain.update(fetch_now=True)
        for extractor_class in EXTRACTOR_CLASSES:
            if hasattr(extractor_class, 'update_dependencies'):
                extractor_class.update_dependencies(self.options)



import logging
import multiprocessing
import os
import signal
import socket
import tempfile
import time
from datetime import datetime
from multiprocessing.connection import wait

import psycopg2

from privacyscanner.exceptions import RetryScan, RescheduleLater
from privacyscanner.filehandlers import NoOpFileHandler
from privacyscanner.jobqueue import JobQueue
from privacyscanner.raven import has_raven, raven
from privacyscanner.result import Result
from privacyscanner.scanmeta import ScanMeta
from privacyscanner.scanmodules import load_modules
from privacyscanner.loghandlers import WorkerWritePipeHandler, ScanStreamHandler
from privacyscanner.utils import kill_everything


_JOB_STARTED_QUERY = """
UPDATE scanner_scaninfo
SET scan_host = %s,
    time_started = %s,
    num_tries = num_tries + 1
WHERE scan_id = %s AND scan_module = %s
"""

_JOB_FINISHED_QUERY = """
UPDATE scanner_scaninfo
SET time_finished = %s
WHERE scan_id = %s AND scan_module = %s
"""

_JOB_FAILED_QUERY = """
UPDATE scanner_scaninfo
SET scan_host = NULL,
    time_started = NULL
WHERE scan_id = %s AND scan_module = %s
"""

_LOG_QUERY = """
INSERT INTO scanner_logentry (scan_id, scan_module, scan_host, time_created, level, message)
VALUES (%s, %s, %s, %s, %s, %s)
"""


class WorkerInfo:
    def __init__(self, worker_id, process, read_pipe, stop_event, ack_event):
        self.id = worker_id
        self.process = process
        self.read_pipe = read_pipe
        self.stop_event = stop_event
        self.ack_event = ack_event
        self.scan_id = None
        self.scan_module = None
        self._heartbeat = None
        self._last_execution_time = None
        self.ping()

    @property
    def pid(self):
        return self.process.pid

    def ping(self):
        self._heartbeat = time.time()

    def ack(self):
        self.ack_event.set()

    def notify_job_started(self, scan_id, scan_module):
        self.scan_id = scan_id
        self.scan_module = scan_module
        self._last_execution_time = time.time()

    def notify_job_finished(self):
        self.scan_id = None
        self.scan_module = None

    notify_job_failed = notify_job_finished

    def get_execution_time(self):
        if self._last_execution_time is None:
            return 0
        return max(time.time() - self._last_execution_time, 0)

    def stop(self):
        self.stop_event.set()

    def __str__(self):
        return '<{}/{} pid={}>'.format(self.scan_id, self.scan_module, self.pid)


class WorkerMaster:
    def __init__(self, db_dsn, scan_module_list, scan_module_options=None,
                 max_tries=3, num_workers=2, max_executions=100,
                 max_execution_times=None, raven_dsn=None):
        self.name = socket.gethostname()
        self._db_dsn = db_dsn
        self.scan_module_list = scan_module_list
        if scan_module_options is None:
            scan_module_options = {}
        self.scan_module_options = scan_module_options
        self.max_tries = max_tries
        self.num_workers = num_workers
        self.max_executions = max_executions
        if max_execution_times is None:
            max_execution_times = {None: None}
        self.max_execution_times = max_execution_times
        self.max_execution_time = max_execution_times.get(None)
        self._raven_dsn = raven_dsn
        self._workers = {}
        self._worker_ids = set(range(num_workers))
        self._terminated_worker_pids = set()
        self._running = False
        self._force_stop = False
        self._conn = None
        self._connect()

    def start(self):
        multiprocessing.set_start_method('spawn')
        signal.signal(signal.SIGINT, self._handle_signal_stop)
        signal.signal(signal.SIGTERM, self._handle_signal_stop)
        signal.signal(signal.SIGUSR1, self._handle_signal_usr1)
        self._running = True
        while self._running:
            self._start_workers()
            self._process_queue()
            self._check_hanging()
            self._remove_workers()
            time.sleep(0.25)
        print('\nGently asking workers to stop after their current job ...')
        for worker_info in self._workers.values():
            worker_info.stop()
        while not self._force_stop and self._workers:
            workers_str = self._get_running_workers_str()
            print('{} workers still alive: {}'.format(len(self._workers), workers_str))
            self._check_hanging()
            self._remove_workers()
            time.sleep(0.25)
        if self._workers:
            print('Forcefully killing workers ...')
            for worker_info in self._workers.values():
                kill_everything(worker_info.pid)
        print('All workers stopped. Shutting down ...')

    def stop(self):
        if self._running:
            self._running = False
        else:
            self._force_stop = True

    def _connect(self):
        if self._conn is None or self._conn.closed:
            self._conn = psycopg2.connect(self._db_dsn)

    def _start_workers(self):
        ppid = os.getpid()
        for i in range(self.num_workers - len(self._workers)):
            worker_id = self._worker_ids.pop()
            stop_event = multiprocessing.Event()
            ack_event = multiprocessing.Event()
            read_pipe, write_pipe = multiprocessing.Pipe(duplex=False)
            args = (worker_id, ppid, self._db_dsn, self.scan_module_list,
                    self.scan_module_options, self.max_tries, self.max_executions,
                    write_pipe, stop_event, ack_event, self._raven_dsn)
            process = WorkerProcess(target=_spawn_worker, args=args)
            process.start()
            worker_info = WorkerInfo(worker_id, process, read_pipe, stop_event, ack_event)
            self._workers[worker_info.pid] = worker_info

    def _process_queue(self):
        while True:
            pipes = [worker_info.read_pipe for worker_info in self._workers.values()
                     if worker_info.process.is_alive()]
            ready_pipes = wait(pipes, timeout=0.1)
            if not ready_pipes:
                break
            for read_pipe in ready_pipes:
                try:
                    event = read_pipe.recv()
                except EOFError:
                    continue
                self._process_queue_event(event)

    def _process_queue_event(self, event):
        pid, action, args = event
        worker_info = self._workers[pid]
        worker_info.ping()
        if action == 'job_started':
            scan_id, scan_module_name, time_started, num_tries = args
            self._event_job_started(scan_id, scan_module_name, time_started)
            worker_info.notify_job_started(scan_id, scan_module_name)
        elif action == 'job_finished':
            self._event_job_finished(
                worker_info.scan_id, worker_info.scan_module, time_finished=args[0])
            worker_info.notify_job_finished()
        elif action == 'job_failed':
            self._event_job_failed(worker_info.scan_id, worker_info.scan_module)
            worker_info.notify_job_failed()
        elif action == 'log':
            log_time, level, message = args
            self._event_job_log(worker_info.scan_id, worker_info.scan_module,
                                log_time, level, message)
        elif action == 'add_file':
            pass
        elif action == 'add_debug_file':
            pass
        worker_info.ack()

    def _event_job_started(self, scan_id, scan_module_name, time_started):
        params = (self.name, time_started, scan_id, scan_module_name)
        self._execute_sql_autocommit(_JOB_STARTED_QUERY, params)

    def _event_job_finished(self, scan_id, scan_module_name, time_finished):
        params = (time_finished, scan_id, scan_module_name)
        self._execute_sql_autocommit(_JOB_FINISHED_QUERY, params)

    def _event_job_failed(self, scan_id, scan_module_name):
        params = (scan_id, scan_module_name)
        self._execute_sql_autocommit(_JOB_FAILED_QUERY, params)

    def _event_job_log(self, scan_id, scan_module_name, log_time, level, message):
        log_time = datetime.fromtimestamp(log_time)
        params = (scan_id, scan_module_name, self.name, log_time, level, message)
        self._execute_sql_autocommit(_LOG_QUERY, params)

    def _execute_sql_autocommit(self, query, params):
        while True:
            try:
                self._connect()
                with self._conn.cursor() as c:
                    c.execute(query, params)
                self._conn.commit()
                break
            except psycopg2.OperationalError:
                print('Database operational error. Retrying after 10 seconds.')
                time.sleep(10)

    def _check_hanging(self):
        for worker_info in self._workers.values():
            max_execution_time = self.max_execution_times.get(
                worker_info.scan_module, self.max_execution_time)
            if max_execution_time is None:
                continue
            if worker_info.get_execution_time() > max_execution_time:
                worker_info.notify_job_failed()
                self._event_job_failed(worker_info.scan_id, worker_info.scan_module)
                kill_everything(worker_info.pid)
                self._terminated_worker_pids.add(worker_info.pid)

    def _remove_workers(self):
        for worker_info in self._workers.values():
            if not worker_info.process.is_alive():
                self._terminated_worker_pids.add(worker_info.pid)
        for pid in self._terminated_worker_pids:
            self._worker_ids.add(self._workers[pid].id)
            del self._workers[pid]
        self._terminated_worker_pids.clear()

    def _handle_signal_stop(self, signum, frame):
        assert signum in (signal.SIGINT, signal.SIGTERM)
        self.stop()

    def _handle_signal_usr1(self, signum, frame):
        assert signum == signal.SIGUSR1
        print('Running workers: {}'.format(self._get_running_workers_str()))

    def _get_running_workers_str(self):
        return ' '.join(str(worker_info) for worker_info in self._workers.values())


def _spawn_worker(*args, **kwargs):
    w = Worker(*args, **kwargs)
    w.run()


class Worker:
    def __init__(self, worker_id, ppid, db_dsn, scan_module_list, scan_module_options,
                 max_tries, max_executions, write_pipe, stop_event, ack_event,
                 raven_dsn):
        self._id = worker_id
        self._pid = os.getpid()
        self._ppid = ppid
        self._max_executions = max_executions
        self._write_pipe = write_pipe
        self._stop_event = stop_event
        self._ack_event = ack_event
        self._old_sigterm = signal.SIG_DFL
        self._old_sigint = signal.SIG_DFL
        self._raven_client = None
        if has_raven and raven_dsn:
            self._raven_client = raven.Client(raven_dsn)
        scan_modules = load_modules(scan_module_list, scan_module_options)
        self._job_queue = JobQueue(db_dsn, scan_modules, max_tries)

    def run(self):
        while self._max_executions > 0:
            # Stop if our master died.
            if self._ppid != os.getppid():
                break

            # Our master asked us to stop. We must obey.
            if self._stop_event.is_set():
                break
            job = self._job_queue.get_job_nowait()
            if job is None:
                time.sleep(1)
                continue
            start_info = (job.scan_id, job.scan_module.name, datetime.today(), job.num_tries)
            self._notify_master('job_started', start_info)
            result = Result(job.current_result, NoOpFileHandler())
            logger = logging.Logger(job.scan_module.name)
            logger.addHandler(WorkerWritePipeHandler(self._pid, self._write_pipe))
            logger.addHandler(ScanStreamHandler())
            scan_meta = ScanMeta(worker_id=self._id, num_tries=job.num_tries)
            with tempfile.TemporaryDirectory() as temp_dir:
                old_cwd = os.getcwd()
                os.chdir(temp_dir)
                try:
                    job.scan_module.logger = logger
                    job.scan_module.scan_site(result, scan_meta)
                except RetryScan:
                    self._job_queue.report_failure()
                    self._notify_master('job_failed', (datetime.today(),))
                except RescheduleLater as e:
                    self._job_queue.reschedule(e.not_before)
                    self._job_queue.report_result(result.get_updates())
                    self._notify_master('job_finished', (datetime.today(),))
                except Exception:
                    logger.exception('Scan module `%s` failed.', job.scan_module.name)
                    self._job_queue.report_failure()
                    self._notify_master('job_failed', (datetime.today(),))
                    if self._raven_client:
                        self._raven_client.captureException(tags={
                            'scan_id': job.scan_id,
                            'scan_module_name': job.scan_module.name
                        }, extra={'result': result.get_results()})
                else:
                    self._job_queue.report_result(result.get_updates())
                    self._notify_master('job_finished', (datetime.today(),))
                finally:
                    os.chdir(old_cwd)
                    kill_everything(self._pid, only_children=True)
            self._max_executions -= 1
        kill_everything(self._pid)

    def _notify_master(self, action, args):
        self._write_pipe.send((self._pid, action, args))
        self._ack_event.wait()
        self._ack_event.clear()


class WorkerProcess(multiprocessing.Process):
    def run(self):
        # We do not want our worker to receive the signals our parent (master)
        # gets. Therefore move it into an own process group.
        os.setpgid(0, 0)
        super().run()



import importlib
import logging
from typing import Any, Dict, List


class ModuleLoadError(Exception):
    pass


class ScanModule:
    name = None  # type: str
    dependencies = None  # type: List[str]
    required_keys = None  # type: List[str]
    logger = None  # type: logging.Logger
    options = None  # type: Dict[str, Any]

    def __init__(self, options):
        self.options = options
        self.logger = logging.Logger(self.name)

    def scan_site(self, result, meta):
        raise NotImplemented

    def update_dependencies(self):
        pass


def load_modules(module_list, module_options):
    scan_modules = {}
    for module_class in module_list:
        module_name, class_name = module_class.rsplit('.', 1)
        try:
            mod = importlib.import_module(module_name)
            class_obj = getattr(mod, class_name)
        except (ImportError, AttributeError) as e:
            raise ModuleLoadError('Could not load module {}.'.format(module_class)) from e
        for attr_name in ('name', 'scan_site', 'dependencies', 'required_keys'):
            if not hasattr(class_obj, attr_name):
                raise ModuleLoadError('Module {} has no attribute `{}`.'.format(
                    module_class, attr_name
                ))
        # Copy options which are for all scan modules to each individual
        # scan module if they are not set, i.e., you can override them
        # for an individual scan module.
        options = module_options.get(class_obj.name, {})
        for key, value in module_options['__all__'].items():
            if key not in options:
                options[key] = value

        scan_modules[class_obj.name] = class_obj(options)
    return scan_modules



from urllib.parse import parse_qs

from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import JavaScriptError, javascript_evaluate


TRACKER_JS = """
(function() {
    let info = {
        'has_ga_object': false,
        'has_gat_object': false,
        'trackers': []
    };
    if (typeof(ga) === 'undefined' && typeof(_gat) === 'undefined') {
        return info;
    }
    
    if (typeof(ga) !== 'undefined') {
        ga(function() {
            info['has_ga_object'] = true;
            ga.getAll().forEach(function(tracker) {
                let anonymize_ip = tracker.get('anonymizeIp');
                info['trackers'].push({
                    'name': tracker.get('name'),
                    'tracking_id': tracker.get('trackingId'),
                    'anonymize_ip': typeof(anonymize_ip) !== 'undefined' ? !!anonymize_ip : false
                });
            });
        });
    }
    if (typeof(_gat) !== 'undefined') {
        info['has_gat_object'] = true;
        _gat._getTrackers().forEach(function(tracker) {
            // There is only a global anonymize in the old _gat API
            let anonymize_ip = _gat.w; 
            info['trackers'].push({
                'name': tracker._getName(),
                'tracking_id': tracker._getAccount(),
                'anonymize_ip': typeof(anonymize_ip) !== 'undefined' ? !!anonymize_ip : false
            });
        });
    }
    
    return info;
})()
""".lstrip()


class GoogleAnalyticsExtractor(Extractor):
    def extract_information(self):
        ga = {
            'has_ga_object': None,
            'has_gat_object': None,
            'trackers': []
        }

        if not self.options['disable_javascript']:
            ga['has_gat_object'] = False
            ga['has_gat_object'] = False
            try:
                info = javascript_evaluate(self.page.tab, TRACKER_JS)
                ga.update(info)
            except JavaScriptError:
                pass
        num_requests_aip = 0
        num_requests_no_aip = 0
        has_ga_requests = False
        for request in self.page.request_log:
            if self._is_google_request(request['parsed_url']):
                if self._is_anonymized(request):
                    num_requests_aip += 1
                else:
                    num_requests_no_aip += 1
                has_ga_requests = True
        ga['has_requests'] = has_ga_requests

        has_ga_js = ga['has_ga_object'] or ga['has_gat_object']
        ga['has_ga_js'] = has_ga_js

        if has_ga_requests or has_ga_js:
            all_set_js = None
            any_set_js = None
            if ga['trackers']:
                trackers_anonymize = [tracker.get('anonymize_ip') for tracker in ga['trackers']]
                if all(isinstance(aip, bool) for aip in trackers_anonymize):
                    all_set_js = all(trackers_anonymize)
                    any_set_js = any(trackers_anonymize)
            aip_ineffective = None
            if all_set_js is not None and any_set_js is not None:
                aip_ineffective = ((any_set_js and not all_set_js) or
                                   (all_set_js and num_requests_no_aip > 0))
            ga['anonymize'] = {
                'all_set_js': all_set_js,
                'any_set_js': any_set_js,
                'num_requests_aip': num_requests_aip,
                'num_requests_no_aip': num_requests_no_aip,
                'aip_ineffective': aip_ineffective
            }
        if not (ga['has_ga_object'] or ga['has_gat_object']):
            del ga['trackers']

        self.result['google_analytics'] = ga

    @staticmethod
    def _is_google_request(parsed_url):
        # Google uses stats.g.doubleclick.net for customers that have
        # enabled the Remarketing with Google Analytics feature.
        ga_domains = ('www.google-analytics.com', 'ssl.google-analytics.com',
                      'stats.g.doubleclick.net')
        if parsed_url.netloc in ga_domains:
            return any(p in parsed_url.path for p in ('collect', '__utm.gif'))

    @staticmethod
    def _is_anonymized(request):
        # There could be conflicting aip options, e.g., when a POST request
        # contains aip=0 in their post data, but aip=1 in the URL.
        # In this case, post data takes precedence.
        aip = None
        if request['method'] == 'POST' and request['post_data']:
            qs = parse_qs(request['post_data'])
            aip = qs.get('aip')
        if aip is None:
            qs = parse_qs(request['parsed_url'].query)
            aip = qs.get('aip')
        if aip and aip[-1] in ('1', 'true'):
            return True
        return False



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import parse_domain


class CookieStatsExtractor(Extractor):
    long_cookie_time = 24 * 60 * 60

    def extract_information(self):
        stats = {}
        for party in ('first', 'third'):
            for duration in ('short', 'long'):
                stats['{}_party_{}'.format(party, duration)] = 0
        cookietrackers = set()
        for cookie in self.result['cookies']:
            prefix = 'third' if cookie['is_thirdparty'] else 'first'
            suffix = 'long' if cookie['lifetime'] > self.long_cookie_time else 'short'
            stats['{}_party_{}'.format(prefix, suffix)] += 1
            if cookie['is_tracker']:
                tracker = parse_domain(cookie['domain'])
                cookietrackers.add(tracker.registered_domain)
        stats['trackers'] = list(sorted(cookietrackers))
        self.result['cookiestats'] = stats



import base64

from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


INSTRUMENTATION_JS = """
function instrumentFunction(func, name, log_type) {
    return function() {
        let retval = func.apply(this, arguments);
        log(log_type, {
            'type': 'function',
            'name': name,
            'arguments': Array.prototype.slice.call(arguments),
            'retval': retval
        });
    }
}

function instrumentProperty(obj, prop, name, log_type) {
    let prototype = obj;
    let descriptor;
    do {
        descriptor = Object.getOwnPropertyDescriptor(prototype, prop);
        if (typeof(descriptor) !== 'undefined') {
            break
        }
        prototype = Object.getPrototypeOf(prototype);
    } while (prototype !== null);

    if (typeof(descriptor) === 'undefined') {
        return;
    }

    let origGetter = descriptor.get;
    let origSetter = descriptor.set;
    
    Object.defineProperty(obj, prop, {
        get: function() {
            let value = origGetter.apply(this, arguments);
            log(log_type, {
                'type': 'property',
                'name': name,
                'value': value,
                'access': 'get'
            });
            return value;
        },
        set: function() {
            log(log_type, {
                'type': 'property',
                'name': name,
                'value': arguments[0],
                'access': 'set'
            });
            return origSetter.apply(this, arguments);
            
        }
    });
}

function instrumentObject(obj, name, properties, log_type) {
    for (let i = 0; i < properties.length; i++) {
        let prop = properties[i];
        if (typeof(obj[prop]) === 'function') {
            let funcName = name + '.' + prop;
            obj[prop] = instrumentFunction(obj[prop], funcName, log_type);
        } else {
            let propName = name + '.' + prop;
            instrumentProperty(obj, prop, propName, log_type);
        }
    }
}

instrumentObject(window.HTMLCanvasElement.prototype,
                'HTMLCanvasElement',
                ['toDataURL'],
                'fingerprinting:canvas');
instrumentObject(window.CanvasRenderingContext2D.prototype,
                 'CanvasRenderingContext2D',
                 ['fillText', 'strokeText'],
                 'fingerprinting:canvas');
"""


class FingerprintingExtractor(Extractor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._canvas = {'calls': [], 'is_fingerprinting': False}
        self._canvas_call_stack = None
        self._canvas_image = None

    def extract_information(self):
        self.result['fingerprinting'] = {
            'canvas': self._canvas
        }
        self._extract_canvas()

    def register_javascript(self):
        return INSTRUMENTATION_JS

    def receive_log(self, log_type, message, call_stack):
        if log_type == 'fingerprinting:canvas':
            self._receive_canvas_log(message, call_stack)

    def _extract_canvas(self):
        uses_text = False
        text_methods = ('CanvasRenderingContext2D.fillText',
                        'CanvasRenderingContext2D.strokeText')
        for call in self._canvas['calls']:
            if call['method'] in text_methods:
                uses_text = True
                break
        if uses_text and self._canvas_call_stack is not None:
            self._canvas['is_fingerprinting'] = True
            self._canvas['call_stack'] = self._canvas_call_stack
            content = None
            try:
                info_part, data_part = self._canvas_image.split(',', 1)
                if 'base64' in info_part:
                    content = base64.b64decode(data_part)
            except ValueError:
                pass
            if content:
                self.result.add_file('fingerprinting_canvas', content)

    def _receive_canvas_log(self, message, call_stack):
        self._canvas['calls'].append({
            'method': message['name'],
            'arguments': message['arguments']
        })
        if message['name'] == 'HTMLCanvasElement.toDataURL':
            # Our first element is the function that was used to override
            # the original function for logging, i.e., that anonymous
            # function returned by instrumentFunction. So we skip it in our
            # call stack because it provides no value.
            self._canvas_call_stack = call_stack[1:]
            self._canvas_image = message['retval']



import io
import json

from privacyscanner.scanmodules.chromedevtools.utils import parse_domain
from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.utils import download_file, file_is_outdated


HSTS_PRELOAD_URL = 'https://cs.chromium.org/codesearch/f/chromium/src/net/http/transport_security_state_static.json'

_hsts_lookup = None


class HSTSPreloadExtractor(Extractor):
    def extract_information(self):
        global _hsts_lookup

        hsts_preload = {
            'is_ready': False,
            'is_preloaded': False
        }
        self.result['https']['hsts_preload'] = hsts_preload
        self.result.mark_dirty('https')

        if _hsts_lookup is None:
            lookup_file = self.options['storage_path'] / 'hsts.json'
            with lookup_file.open() as f:
                _hsts_lookup = json.load(f)

        domain = parse_domain(self.result['final_url']).registered_domain
        is_preloaded = domain in _hsts_lookup

        # Iterate over all subdomains and check if any of it is preloaded.
        # We have to handle three cases:
        # 1) all subdomains are not in the preload list. Reject.
        # 2) a subdomain is in the preload list, and include_subdomains is set.
        #    Accept in this case.
        # 3) a subdomain is in the preload list, but include_subdomains
        #    is not set. Then we have to do two things. Firstly, continue
        #    searching: maybe another subdomain of the current subdomain
        #    is in the list and has include_subdomains. See case 2. Secondly,
        #    the full domain might be in the lookup. This has already been
        #    checked beforehand, so nothing to do.
        current_domain = ''
        for part in domain.split('.'):
            current_domain = part + '.' + current_domain
            if current_domain in _hsts_lookup:
                include_subdomains = _hsts_lookup[current_domain]
                if include_subdomains:
                    is_preloaded = True
                    break
            else:
                break

        hsts_header = self.result['security_headers']['Strict-Transport-Security']
        if hsts_header is None:
            return

        # There are some big players who got exceptions from the standard
        # requirements to be HSTS ready, therefore we treat them as HSTS ready
        # if they are already in the preload list. However, we require the HSTS
        # header to be set (see return statement above).
        hsts_preload['is_preloaded'] = is_preloaded
        if is_preloaded:
            hsts_preload['is_ready'] = True
            return

        # According to hstspreload.org, these are the criteria for being ready
        # to be included in the HSTS preload list:
        #
        # 1. Serve a valid certificate.
        # 2. Redirect from HTTP to HTTPS on the same host, if you are listening
        #    on port 80.
        # 3. Serve all subdomains over HTTPS.
        #    In particular, you must support HTTPS for the www subdomain if a
        #    DNS record for that subdomain exists.
        # 4. Serve an HSTS header on the base domain for HTTPS requests:
        #    4.1 The max-age must be at least 31536000 seconds (1 year).
        #    4.2 The includeSubDomains directive must be specified.
        #    4.3 The preload directive must be specified.
        #    4.4 If you are serving an additional redirect from your HTTPS site,
        #        that redirect must still have the HSTS header (rather than the
        #        page it redirects to).

        fail_reasons = []
        if not self.result['final_url'].startswith('https://'):
            fail_reasons.append('no-https-redirect')
        if not hsts_header['includeSubDomains']:
            fail_reasons.append('no-include-subdomains')
        if hsts_header['max-age'] is None:
            fail_reasons.append('no-max-age')
        elif hsts_header['max-age'] < 31536000:
            fail_reasons.append('max-age-too-short')
        if not hsts_header['preload']:
            fail_reasons.append('missing-preload')

        fail_reasons.sort()
        hsts_preload['is_ready'] = len(fail_reasons) == 0
        if fail_reasons:
            hsts_preload['fail_reasons'] = fail_reasons

    @classmethod
    def update_dependencies(cls, options):
        lookup_file = options['storage_path'] / 'hsts.json'
        if not file_is_outdated(lookup_file, 3600 * 24 * 7):
            return
        buf = io.BytesIO()
        download_url = options.get('hsts_preload_url', HSTS_PRELOAD_URL)
        download_file(download_url, buf)
        plain_json = ''.join(line for line in buf.getvalue().decode().splitlines()
                             if not line.lstrip().startswith('//'))
        hsts_data = json.loads(plain_json)
        lookup = {entry['name']: entry.get('include_subdomains')
                  for entry in hsts_data['entries']}
        with lookup_file.open('w') as f:
            json.dump(lookup, f)



from typing import NamedTuple

import psycopg2
from psycopg2.extras import Json


_FETCH_JOB_QUERY = """
WITH job AS (
  SELECT sj1.id, si.num_tries
  FROM scanner_scanjob AS sj1,
       scanner_scaninfo AS si
  WHERE NOT EXISTS ( -- Make sure our dependencies are processed first
    SELECT id
    FROM scanner_scanjob AS sj2
    WHERE sj2.dependency_order < sj1.dependency_order AND
          sj2.scan_id = sj1.scan_id
  ) AND sj1.scan_module IN %s
    AND si.scan_id = sj1.scan_id
    AND si.scan_module = sj1.scan_module
    AND si.num_tries < %s
    AND (sj1.not_before IS NULL OR sj1.not_before <= NOW())
  ORDER BY sj1.priority DESC, sj1.scan_id, sj1.dependency_order
  FOR UPDATE OF sj1 SKIP LOCKED
  LIMIT 1
)
DELETE FROM scanner_scanjob
WHERE id = (SELECT id FROM job)
RETURNING id, scan_id, scan_module, (SELECT num_tries FROM job) AS num_tries, dependency_order, priority
"""

_FETCH_RESULT_QUERY = """
SELECT (kv).key, (kv).value
FROM (
  SELECT jsonb_each(result) AS kv
  FROM scanner_scan
  WHERE id = %s
) AS s
WHERE (kv).key IN %s
"""

_UPDATE_RESULT_QUERY = """
UPDATE scanner_scan
SET result = result || %s::jsonb
WHERE id = %s
"""

_RESCHEDULE_JOB_QUERY = """
INSERT INTO scanner_scanjob
(scan_module, priority, dependency_order, scan_id, not_before)
VALUES (%s, %s, %s, %s, %s)
"""

_INCREASE_TRIES_QUERY = """
UPDATE scanner_scaninfo
SET num_tries = GREATEST(0, num_tries - 1)
WHERE scan_id = %s AND scan_module = %s
"""

class Job(NamedTuple):
    scan_id: int
    scan_module: object
    current_result: dict
    num_tries: int
    dependency_order: int
    priority: int


class JobQueue:
    def __init__(self, dsn, scan_modules, max_tries):
        self._dsn = dsn
        self._scan_modules = scan_modules
        self._available_modules = tuple(self._scan_modules.keys())
        self._max_tries = max_tries
        self._last_job = None
        self._conn = None
        self._connect()

    def report_result(self, updates):
        assert self._last_job is not None
        with self._conn.cursor() as c:
            c.execute(_UPDATE_RESULT_QUERY, (Json(updates), self._last_job.scan_id))
        self._last_job = None
        self._conn.commit()

    def report_failure(self):
        assert self._last_job is not None
        self._last_job = None
        self._conn.rollback()

    def _connect(self):
        self._conn = psycopg2.connect(self._dsn)

    def get_job_nowait(self):
        assert self._last_job is None
        if self._conn.closed:
            self._connect()
        with self._conn.cursor() as c:
            c.execute(_FETCH_JOB_QUERY, (self._available_modules, self._max_tries))
            job = c.fetchone()
            if job:
                job_id, scan_id, scan_module_name, num_tries, dependency_order, priority = job
                scan_module = self._scan_modules[scan_module_name]
                if scan_module.required_keys:
                    c.execute(_FETCH_RESULT_QUERY, (scan_id, tuple(scan_module.required_keys)))
                    result = dict(c.fetchall())
                else:
                    result = {}
                job = Job(scan_id, scan_module, result, num_tries, dependency_order, priority)
                self._last_job = job
                return job

    def reschedule(self, not_before=None):
        assert self._last_job is not None
        with self._conn.cursor() as c:
            params = (self._last_job.scan_module.name, self._last_job.priority,
                      self._last_job.dependency_order, self._last_job.scan_id,
                      not_before)
            c.execute(_RESCHEDULE_JOB_QUERY, params)
            c.execute(_INCREASE_TRIES_QUERY, (self._last_job.scan_id,
                                              self._last_job.scan_module.name))
